@@ -54,22 +54,10 @@ static const size_t MAX_UNRECOGNIZED_LINES = 1000;
 static const size_t INDEX_RESERVE_INCREMENT = 1024;
 
 logfile::logfile(const string &filename, logfile_open_options &loo)
-throw (error)
-    : lf_filename(filename),
-      lf_index_time(0),
-      lf_index_size(0),
-      lf_sort_needed(false),
-      lf_is_closed(false),
-      lf_logline_observer(NULL),
-      lf_logfile_observer(NULL),
-      lf_longest_line(0),
-      lf_text_format(TF_UNKNOWN)
+    : lf_filename(filename)
 {
-    require(filename.size() > 0);
+    require(!filename.empty());
 
-    this->lf_time_offset.tv_sec = 0;
-    this->lf_time_offset.tv_usec = 0;
-    
     memset(&this->lf_stat, 0, sizeof(this->lf_stat));
     if (loo.loo_fd == -1) {
         char resolved_path[PATH_MAX];
@@ -119,7 +107,7 @@ logfile::~logfile()
 {
 }
 
-bool logfile::exists(void) const
+bool logfile::exists() const
 {
     struct stat st;
 
@@ -146,15 +134,19 @@ void logfile::set_format_base_time(log_format *lf)
     lf->lf_date_time.set_base_time(file_time);
 }
 
-bool logfile::process_prefix(off_t offset, shared_buffer_ref &sbr)
+bool logfile::process_prefix(off_t offset,
+                             shared_buffer_ref &sbr,
+                             const line_value &lv)
 {
     log_format::scan_result_t found = log_format::SCAN_NO_MATCH;
     size_t prescan_size = this->lf_index.size();
     time_t prescan_time = 0;
     bool retval = false;
 
-    if (this->lf_format.get() != NULL) {
-        prescan_time = this->lf_index[0].get_time();
+    if (this->lf_format.get() != nullptr) {
+        if (!this->lf_index.empty()) {
+            prescan_time = this->lf_index[0].get_time();
+        }
         /* We've locked onto a format, just use that scanner. */
         found = this->lf_format->scan(this, this->lf_index, offset, sbr);
     }
@@ -211,6 +203,9 @@ bool logfile::process_prefix(off_t offset, shared_buffer_ref &sbr)
 
     switch (found) {
         case log_format::SCAN_MATCH:
+            if (!this->lf_index.empty()) {
+                this->lf_index.back().set_valid_utf(lv.lv_valid_utf);
+            }
             if (!this->lf_index.empty() && prescan_time != this->lf_index[0].get_time()) {
                 retval = true;
             }
@@ -220,14 +215,7 @@ bool logfile::process_prefix(off_t offset, shared_buffer_ref &sbr)
 
                 if (latest < second_to_last) {
                     if (this->lf_format->lf_time_ordered) {
-                        log_debug(
-                            "%s:%d: out-of-time-order line detected %d.%03d < %d.%03d",
-                            this->lf_filename.c_str(),
-                            prescan_size,
-                            latest.get_time(),
-                            latest.get_millis(),
-                            second_to_last.get_time(),
-                            second_to_last.get_millis());
+                        this->lf_out_of_time_order_count += 1;
                         for (size_t lpc = prescan_size;
                              lpc < this->lf_index.size(); lpc++) {
                             logline &line_to_update = this->lf_index[lpc];
@@ -244,7 +232,7 @@ bool logfile::process_prefix(off_t offset, shared_buffer_ref &sbr)
             }
             break;
         case log_format::SCAN_NO_MATCH: {
-            logline::level_t last_level = logline::LEVEL_UNKNOWN;
+            log_level_t last_level = LEVEL_UNKNOWN;
             time_t last_time = this->lf_index_time;
             short last_millis = 0;
             uint8_t last_mod = 0, last_opid = 0;
@@ -259,18 +247,19 @@ bool logfile::process_prefix(off_t offset, shared_buffer_ref &sbr)
                 last_time = ll.get_time();
                 last_millis = ll.get_millis();
                 if (this->lf_format.get() != NULL) {
-                    last_level = (logline::level_t)(ll.get_level_and_flags() |
-                        logline::LEVEL_CONTINUED);
+                    last_level = (log_level_t)(ll.get_level_and_flags() |
+                        LEVEL_CONTINUED);
                 }
                 last_mod = ll.get_module_id();
                 last_opid = ll.get_opid();
             }
-            this->lf_index.push_back(logline(offset,
-                                             last_time,
-                                             last_millis,
-                                             last_level,
-                                             last_mod,
-                                             last_opid));
+            this->lf_index.emplace_back(offset,
+                                        last_time,
+                                        last_millis,
+                                        last_level,
+                                        last_mod,
+                                        last_opid);
+            this->lf_index.back().set_valid_utf(lv.lv_valid_utf);
             break;
         }
         case log_format::SCAN_INCOMPLETE:
@@ -281,7 +270,6 @@ bool logfile::process_prefix(off_t offset, shared_buffer_ref &sbr)
 }
 
 logfile::rebuild_result_t logfile::rebuild_index()
-throw (line_buffer::error, logfile::error)
 {
     rebuild_result_t retval = RR_NO_NEW_LINES;
     struct stat st;
@@ -304,13 +292,15 @@ throw (line_buffer::error, logfile::error)
 
         // We haven't reached the end of the file.  Note that we use the
         // line buffer's notion of the file size since it may be compressed.
-        bool has_format = this->lf_format.get() != NULL;
+        bool has_format = this->lf_format.get() != nullptr;
         struct rusage begin_rusage;
         shared_buffer_ref sbr;
         off_t  last_off, off;
         line_value lv;
         size_t begin_size = this->lf_index.size();
         bool record_rusage = this->lf_index.size() == 1;
+        off_t begin_index_size = this->lf_index_size;
+        size_t rollback_size = 0;
 
         if (record_rusage) {
             getrusage(RUSAGE_SELF, &begin_rusage);
@@ -325,8 +315,10 @@ throw (line_buffer::error, logfile::error)
              */
             while (this->lf_index.back().get_sub_offset() != 0) {
                 this->lf_index.pop_back();
+                rollback_size += 1;
             }
             this->lf_index.pop_back();
+            rollback_size += 1;
 
             this->lf_line_buffer.clear();
             if (!this->lf_index.empty()) {
@@ -346,7 +338,7 @@ throw (line_buffer::error, logfile::error)
         }
         last_off = off;
         if (this->lf_logline_observer != NULL) {
-            this->lf_logline_observer->logline_restart(*this);
+            this->lf_logline_observer->logline_restart(*this, rollback_size);
         }
 
         bool sort_needed = this->lf_sort_needed;
@@ -355,6 +347,8 @@ throw (line_buffer::error, logfile::error)
         while (this->lf_line_buffer.read_line(off, sbr, &lv)) {
             size_t old_size = this->lf_index.size();
 
+            // Update this early so that line_length() works
+            this->lf_index_size = off;
             if (old_size == 0) {
                 shared_buffer_ref avail_sbr;
 
@@ -365,37 +359,37 @@ throw (line_buffer::error, logfile::error)
 
             this->lf_longest_line = std::max(this->lf_longest_line, sbr.length());
             this->lf_partial_line = lv.lv_partial;
-            sort_needed = this->process_prefix(last_off, sbr) || sort_needed;
+            sort_needed = this->process_prefix(last_off, sbr, lv) || sort_needed;
             last_off = off;
 
             if (old_size > this->lf_index.size()) {
                 old_size = 0;
             }
 
-            for (logfile::iterator iter = this->begin() + old_size;
+            for (auto iter = this->begin() + old_size;
                     iter != this->end(); ++iter) {
-                if (this->lf_logline_observer != NULL) {
+                if (this->lf_logline_observer != nullptr) {
                     this->lf_logline_observer->logline_new_line(*this, iter, sbr);
                 }
             }
 
-            if (this->lf_logfile_observer != NULL) {
+            if (this->lf_logfile_observer != nullptr) {
                 this->lf_logfile_observer->logfile_indexing(
                     *this,
                     this->lf_line_buffer.get_read_offset(off),
                     st.st_size);
             }
 
-            if (!has_format && this->lf_format.get() != NULL) {
+            if (!has_format && this->lf_format != nullptr) {
                 break;
             }
         }
 
-        if (this->lf_logline_observer != NULL) {
+        if (this->lf_logline_observer != nullptr) {
             this->lf_logline_observer->logline_eof(*this);
         }
 
-        if (record_rusage && (off - this->lf_index_size) > (500 * 1024)) {
+        if (record_rusage && (off - begin_index_size) > (500 * 1024)) {
             struct rusage end_rusage;
 
             getrusage(RUSAGE_SELF, &end_rusage);
@@ -427,6 +421,13 @@ throw (line_buffer::error, logfile::error)
         this->lf_index_time = st.st_mtime;
     }
 
+    if (this->lf_out_of_time_order_count) {
+        log_info("Detected %d out-of-time-order lines in file: %s",
+                 this->lf_out_of_time_order_count,
+                 this->lf_filename.c_str());
+        this->lf_out_of_time_order_count = 0;
+    }
+
     return retval;
 }
 
@@ -437,15 +438,21 @@ void logfile::read_line(logfile::iterator ll, string &line_out)
         shared_buffer_ref sbr;
 
         line_out.clear();
-        if (this->lf_line_buffer.read_line(off, sbr)) {
-            if (this->lf_format.get() != NULL) {
-                this->lf_format->get_subline(*ll, sbr);
+
+        size_t line_len = this->line_length(ll, false);
+
+        if (ll->is_valid_utf()) {
+            if (!this->lf_line_buffer.read_range(off, line_len, sbr)) {
+                return;
             }
-            line_out.append(sbr.get_data(), sbr.length());
+        } else if (!this->lf_line_buffer.read_line(off, sbr)) {
+            return;
         }
-        else {
-            /* XXX */
+
+        if (this->lf_format != nullptr) {
+            this->lf_format->get_subline(*ll, sbr);
         }
+        line_out.append(sbr.get_data(), sbr.length());
     }
     catch (line_buffer::error & e) {
         /* ... */
@@ -456,13 +463,20 @@ bool logfile::read_line(logfile::iterator ll, shared_buffer_ref &sbr)
 {
     try {
         off_t       off = ll->get_offset();
+        size_t line_len = this->line_length(ll, false);
 
-        if (this->lf_line_buffer.read_line(off, sbr)) {
-            if (this->lf_format.get() != NULL) {
-                this->lf_format->get_subline(*ll, sbr);
+        if (ll->is_valid_utf()) {
+            if (!this->lf_line_buffer.read_range(off, line_len, sbr)) {
+                return false;
             }
-            return true;
+        } else if (!this->lf_line_buffer.read_line(off, sbr)) {
+            return false;
         }
+
+        if (this->lf_format != nullptr) {
+            this->lf_format->get_subline(*ll, sbr);
+        }
+        return true;
     }
     catch (line_buffer::error & e) {
     }
@@ -477,7 +491,7 @@ void logfile::read_full_message(logfile::iterator ll,
 
     do {
         try {
-            off_t       off = ll->get_offset();
+            off_t off = ll->get_offset();
             shared_buffer_ref sbr;
 
             if (stream.tellp() > 0) {
@@ -516,7 +530,7 @@ void logfile::read_full_message(logfile::iterator ll,
         off_t       off = ll->get_offset();
 
         if (this->lf_line_buffer.read_range(off, line_len, msg_out)) {
-            if (this->lf_format.get() != NULL) {
+            if (this->lf_format.get() != nullptr) {
                 this->lf_format->get_subline(*ll, msg_out, true);
             }
         }
@@ -532,7 +546,7 @@ void logfile::read_full_message(logfile::iterator ll,
 void logfile::set_logline_observer(logline_observer *llo)
 {
     this->lf_logline_observer = llo;
-    if (llo != NULL) {
+    if (llo != nullptr) {
         this->reobserve_from(this->begin());
     }
 }
@@ -560,3 +574,8 @@ void logfile::reobserve_from(iterator iter)
         this->lf_logline_observer->logline_eof(*this);
     }
 }
+
+filesystem::path logfile::get_path() const
+{
+    return this->lf_filename;
+}
